/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define RCI_INTERNAL
#ifndef __RCI__
#include "rci.h"
#endif

// this is necessary for the Paeth filter
int PaethPredictor( int left, int above, int aboveleft )
{
	int p, pa, pb, pc;

	p = left + above - aboveleft;
	pa = p - left;
	if( pa < 0 ) pa = -pa;
	pb = p - above;
	if( pb < 0 ) pb = -pb;
	pc = p - aboveleft;
	if( pc < 0 ) pc = -pc;
	if( pa <= pb && pa <= pc )
		return left;
	else if( pb <= pc )
		return above;
	else
		return aboveleft;
}

void RciCalcHist(RciUns8 *ptr, RciUns32 len, RciUns32 *hist)
{
	while(len--)
		hist[*ptr++]++;
}

RciUns8 RciFindMedian(RciUns8 *ptr, RciUns32 len)
{
	RciUns32 hist[256];
	RciUns8 best;
	RciInt16 i;

	RciMemSet((void *)hist,0,sizeof(RciUns32)*256);
	RciCalcHist(ptr,len,hist);
	
	best = 0;
	for(i=1;i<256;i++)
		if( hist[i] > hist[best] )
			best = (RciUns8)i;
	return best;
}

#define MAX(a,b) ( (a)>(b) ? (a) : (b) )
#define MIN(a,b) ( (a)<(b) ? (a) : (b) )

static int sTable[256];
static int sTableInit = 0;

// this filters a tile and returns the filter type used
RciUns8 RciFilterTile( int filter, RciUns8 *tile, RciUns16 planeDepth, RciUns32 rows, RciUns32 cols,
	RciUns8 *filterParam)
{

	RciUns32 total;
	double noAvg, subAvg, paethAvg, upAvg, avgAvg, offsetAvg, locoAvg, lowestAvg;
	double subScale, paethScale, upScale, avgScale, offsetScale, locoScale;
	RciInt32 i,j;
	int left, curr, temp;
	int above, aboveleft, aboveright;
	RciUns8 *cp;
	RciUns8 avgValue; // average pixel value.
	RciUns8 selectedFilter = filter;

	// we do all of them on a byte basis
	cols *= planeDepth>>3;
	
	// This opt seems to help enough to be worth it but not hugely.
	if( !sTableInit )
	{
		sTableInit = 1;
		for(i=0;i<256;i++)
			sTable[i] = i < 128 ? i : 256 - i;
	}

	// This could be changed to scale the weights if desired.
	subScale = paethScale = upScale = avgScale = offsetScale = locoScale = 1.0;

	// run each filter without storing the results, keeping an average
	// of the absolute value of the result of each, and use that to
	// decide which one.

	// Try to autodetermine the best filter to use if one is not specified.
	if( ! selectedFilter )
	{
		// find hueristic for no filter
		total = 0;
		cp = tile;
		i = rows * cols;
		while(i--)
		{
			// total += *cp < 128 ? *cp : 256 - *cp;
			total += sTable[*cp];
			cp++;
		}
		noAvg = (double)total;
		noAvg /= (double)(rows*cols);
		
		//printf("noAvg is %g\n",noAvg);

		// find heuristic for sub filter
		total = 0;
		left = 0; // startup condition
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				curr = tile[j*cols+i];
				temp = curr - left;
				
				// save stats.
				if( temp < 0 )
					temp += 256;
				// total += temp < 128 ? temp : 256 - temp;
				total += sTable[temp];
				
				left = curr;
			}
		}
		subAvg = (double)total;
		subAvg /= (double)(rows*cols);

		//printf("subAvg is %g\n",subAvg);
		
		// find hueristic for Paeth filter
		total = 0;
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
				}
				else
					aboveleft = above = 0;

				left = i ? tile[j*cols+i-1] : 0;

				curr = tile[j*cols+i];
				temp = curr - PaethPredictor(left,above,aboveleft);
				if( temp < 0 )
					temp += 256;
				//total += temp < 0 ? -temp : temp;
				// total += temp < 128 ? temp : 256 - temp;
				total += sTable[temp];
			}
			
		}
		paethAvg = (double)total;
		paethAvg /= (double)(rows*cols);
		//printf("paethAvg is %g\n",paethAvg);
		
		// find heuristic for up filter
		total = 0;
		above = 0;
		for(i=0;i<(RciInt32)cols;i++)
		{
			for(j=0;j<(RciInt32)rows;j++)
			{
				curr = tile[j*cols+i];
				temp = curr - above;
				if( temp < 0 )
					temp += 256;
				// total += temp < 128 ? temp : 256 - temp;
				total += sTable[temp];

				above = curr;
			}
		}
		upAvg = (double)total;
		upAvg /= (double)(rows*cols);
		
		// find heuristic for avg filter
		total = 0;
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				above = j ? tile[(j-1)*cols+i] : 0;

				left = i ? tile[j*cols+i-1] : 0;

				curr = tile[j*cols+i];
				temp = curr - ( (above+left)>>1 );

				if( temp < 0 )
					temp += 256;

				// total += temp < 128 ? temp : 256 - temp;
				total += sTable[temp];
			}
		}
		avgAvg = (double)total;
		avgAvg /= (double)(rows*cols);
		
		// find heuristic for offset filter.
		// average seems to suck cock, let's try median
		avgValue = RciFindMedian(tile,rows*cols);

		total = 0;
		cp = tile;
		i = rows * cols;
		while(i--)
		{
			temp = *cp++ - avgValue;
			if( temp < 0 )
				temp += 256;
			// total += temp < 128 ? temp : 256 - temp;
			total += sTable[temp];
		}
		offsetAvg = (double)total;
		offsetAvg /= (double)rows*cols;

		// find heuristic for Loco filter
		total = 0;
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
					aboveright = (i == cols-1) ? 0 : tile[(j-1)*cols+i+1];
				}
				else
					aboveright = aboveleft = above = 0;
				
				left = i ? tile[j*cols+i-1] : 0;
				
				// c b d
				// a x
				if( aboveleft < MAX(left,above) )
					temp = MIN(left,above);
				else if( aboveleft < MIN(left,above) )
					temp = MAX(left,above);
				else
					temp = left + above - aboveleft;

				curr = tile[j*cols+i];
				temp = curr - temp;
				if( temp < 0 )
					temp += 256;
				// total += temp < 128 ? temp : 256 - temp;
				total += sTable[temp];
			}
		}
		locoAvg = (double)total;
		locoAvg /= (double)rows*cols;


		// decide which filter to use.

		// scale the averages.
		subAvg *= subScale;
		paethAvg *= paethScale;
		upAvg *= upScale;
		avgAvg *= avgScale;
		offsetAvg *= offsetScale;
		locoAvg *= locoScale;

		//printf("noAvg is %g, paethAvg is %g, locoAvg is %g\n",noAvg,paethAvg,locoAvg);

		// default to none.

		lowestAvg = noAvg;
		selectedFilter = kRciNoFilter;
		
		if( paethAvg < lowestAvg )
		{
			lowestAvg = paethAvg;
			selectedFilter = kRciPaethFilter;
		}
		if( upAvg < lowestAvg )
		{
			lowestAvg = avgAvg;
			selectedFilter = kRciAvgFilter;
		}
		if( subAvg < lowestAvg )
		{
			selectedFilter = kRciSubFilter;
			lowestAvg = subAvg;
		}
		if( upAvg < lowestAvg )
		{
			selectedFilter = kRciUpFilter;
			lowestAvg = upAvg;
		}
		
		if( offsetAvg < lowestAvg )
		{
			// set it to no filter, because we're running this 
			// as a post processor on all of them.
			selectedFilter = kRciNoFilter; // kRciOffsetFilter;
			lowestAvg = offsetAvg;
		}

		if( locoAvg < lowestAvg )
		{
			selectedFilter = kRciLocoFilter;
			lowestAvg = locoAvg;
		}
	}

	if( selectedFilter == kRciSubFilter )
	{

		left = 0; // startup condition
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				curr = tile[j*cols+i];
				temp = curr - left;
				if( temp < 0 )
					temp += 256;
				tile[j*cols+i] = temp;

				left = curr;
			}
		}
	}
	else if( selectedFilter == kRciPaethFilter )
	{
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
				}
				else
					aboveleft = above = 0;
	
				left = i ? tile[j*cols+i-1] : 0;
	
				curr = tile[j*cols+i];
				temp = curr - PaethPredictor(left,above,aboveleft);
				if( temp < 0 )
					temp += 256;
				tile[j*cols+i] = temp;
			}
		}
	}
	else if( selectedFilter == kRciUpFilter )
	{
		above = 0;
		for(i=0;i<(RciInt32)cols;i++)
		{
			for(j=0;j<(RciInt32)rows;j++)
			{
				curr = tile[j*cols+i];
				temp = curr - above;
				if( temp < 0 )
					temp += 256;
				tile[j*cols+i] = temp;

				above = curr;
			}
		}

	}
	else if( selectedFilter == kRciAvgFilter )
	{
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				above = j ? tile[(j-1)*cols+i] : 0;
	
				left = i ? tile[j*cols+i-1] : 0;
	
				curr = tile[j*cols+i];
				temp = curr - ( (above+left)>>1 );
	
				if( temp < 0 )
					temp += 256;
	
				tile[j*cols+i] = temp;
			}
		}
	}
	else if( selectedFilter == kRciLocoFilter )
	{
		for(j=rows-1;j>=0;j--)
		{
			for(i=cols-1;i>=0;i--)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
					aboveright = (i == cols-1) ? 0 : tile[(j-1)*cols+i+1];
				}
				else
					aboveright = aboveleft = above = 0;
				
				left = i ? tile[j*cols+i-1] : 0;
				
				// c b d
				// a x
				curr = tile[j*cols+i];
				if( aboveleft < MAX(left,above) )
					temp = MIN(left,above);
				else if( aboveleft < MIN(left,above) )
					temp = MAX(left,above);
				else
					temp = left + above - aboveleft;
				temp = curr - temp;
				if( temp < 0 )
					temp += 256;

				tile[j*cols+i] = temp;
			}
		}
	}
	// else if( selectedFilter == kRciOffsetFilter )
	// trying this as a post processor to the other filters.
	{
		// store filter param
		avgValue = RciFindMedian(tile,rows*cols);
		*filterParam = avgValue;
		cp = tile;
		i = rows * cols;
		while(i--)
		{
			temp = *cp;
			temp -= avgValue;
			if( temp < 0 )
				temp += 256;
			*cp++ = temp;
		}
	}

	//printf("Selected filter is %d\n",(int)selectedFilter);

	return selectedFilter;
}

// this reverses the filter done on a tile.
void RciInvFilterTile( RciUns8 *tile, RciUns16 planeDepth, RciUns32 rows, RciUns32 cols,
	RciUns8 filterType, RciUns8 filterParam )
{
	RciInt32 i,j;
	RciUns8 *cp;
	int left, above, aboveleft, aboveright, curr, temp;

	// if( filterType == kRciOffsetFilter )
	// trying this on top of all the other filters.
	{
		cp = tile;
		i = rows * cols;
		while(i--)
		{
			temp = *cp;
			temp += filterParam; // stored average value.
			temp &= 0xFF;
			*cp++ = temp;
		}
	}
	//else 
	if( filterType == kRciSubFilter )
	{
		left = 0; // startup condition
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				temp = tile[j*cols+i];
				curr = left + temp;

				if( curr > 256 )
					curr -= 256;

				tile[j*cols+i] = curr;

				left = curr;
			}
		}
	}
	else if( filterType == kRciPaethFilter )
	{
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
				}
				else
					aboveleft = above = 0;
	
				left = i ? tile[j*cols+i-1] : 0;
	
				temp = tile[j*cols+i];
				curr = temp + PaethPredictor(left,above,aboveleft);
				if( curr > 256 )
					curr -= 256;

				tile[j*cols+i] = curr;
			}
		}
	}
	else if( filterType == kRciUpFilter )
	{
		above = 0;
		for(i=0;i<(RciInt32)cols;i++)
		{
			for(j=0;j<(RciInt32)rows;j++)
			{
				temp = tile[j*cols+i];
				curr = temp + above;
				if( curr > 256 )
					curr -= 256;
				tile[j*cols+i] = curr;

				above = curr;
			}
		}
	}
	else if( filterType == kRciAvgFilter )
	{
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				above = j ? tile[(j-1)*cols+i] : 0;
	
				left = i ? tile[j*cols+i-1] : 0;
	
				temp = tile[j*cols+i];
				curr = temp + ( (above+left)>>1 );
	
				if( curr > 256 )
					curr -= 256;

				tile[j*cols+i] = curr;
			}
		}
	}
	else if( filterType == kRciLocoFilter )
	{
		for(j=0;j<(RciInt32)rows;j++)
		{
			for(i=0;i<(RciInt32)cols;i++)
			{
				if( j )
				{
					above = tile[(j-1)*cols+i];
					aboveleft = i ? tile[(j-1)*cols+i-1] : 0;
					aboveright = (i == cols-1) ? 0 : tile[(j-1)*cols+i+1];
				}
				else
					aboveright = aboveleft = above = 0;
				
				left = i ? tile[j*cols+i-1] : 0;
				
				// c b d
				// a x
				if( aboveleft < MAX(left,above) )
					temp = MIN(left,above);
				else if( aboveleft < MIN(left,above) )
					temp = MAX(left,above);
				else
					temp = left + above - aboveleft;

				curr = tile[j*cols+i];
				curr += temp;
				if( curr > 256 )
					curr -= 256;

				tile[j*cols+i] = curr;
			}
		}
	}
}
