/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define RCI_INTERNAL
#ifndef __RCI__
#include "rci.h"
#endif


// The functions in this file read and write an extremely simple
// format I use for testing called RRGB
// it is simply the 8 bytes "RAW RGB " followed by 2 byte big endian order width
// of image then the 2 byte big endian height of image, then all of the 24 bit
// RGB image data in interleaved row major order left to right and top to bottom.

// this reads a RRGB image and fills in the RciState structure to match it
// rci->fd must be open already of course, but this fills in everything else itself.
int ReadRRGBImage(RciState *rci)
{
	int err = 0;
	char cookie[8];
	RciUns16 temp;
	RciUns32 row,rowBytes;

	rci->dataStart = kRciHdrLen; // (char *)&(rci->image) - (char *)rci;
	rci->version = kRciVersion;
	rci->planes = 3;
	rci->colorMode = kRciRGB;
	rci->planeDepth = 8;
	
	err = RciRead(rci->fd,(RciUns8 *)cookie,8);
	if( err )
		return err == kRciEOFErr ? kRciNotFmtErr : err;
	
	if( RciMemCmp(cookie,"RAW RGB ",8) )
	{
		err = kRciNotFmtErr;
		return err;
	}
	
	err = RciRead16BE(rci->fd,&temp);
	if( err )
		return err;
	rci->width = temp;
	err = RciRead16BE(rci->fd,&temp);
	if( err )
		return err;
	rci->height = temp;
	rci->tileHeight = rci->tileWidth = 0;
	
	// read the image data.
	rci->image = RciAlloc( 3 * rci->width * rci->height );
	if( !rci->image )
		return kRciNoMem;
	
	rowBytes = 3*rci->width;
	for(row=0;row<rci->height;row++)
	{
		err = RciRead( rci->fd, ((RciUns8 *)rci->image)+row*rowBytes, rowBytes );
		if( err )
			return err;
	}
	return err;
}

int WriteRRGBImage(RciState *rci)
{
	// This function/format only supports writing 24 bit RGB images.
	int err = 0;
	RciUns16 temp;
	RciUns32 row,rowBytes;

	if(rci->planes != 3 || rci->planeDepth != 8 || rci->colorMode != kRciRGB)
		return kRciStateErr;

	err = RciWrite(rci->fd,(const RciUns8 *)"RAW RGB ",8);
	if(err)
		return err;

	temp = (RciUns16)rci->width;
	err = RciWrite16BE( rci->fd, temp );
	if( err )
		return err;
	
	temp = (RciUns16)rci->height;
	err = RciWrite16BE( rci->fd, temp );
	if( err )
		return err;
	
	// This only deals with 24 bit images.
	rowBytes = 3*rci->width;
	for(row=0;row<rci->height;row++)
	{
		err = RciWrite( rci->fd, ((RciUns8 *)rci->image)+row*rowBytes, rowBytes );
		if( err )
			break;
	}
	return err;
}
