/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2006, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#define RCI_INTERNAL
#ifndef __RCI__
#include "rci.h"
#endif


// calculate the number of tiles per plane
RciUns32 RciNumTiles( RciState *rci)
{
	RciUns32 numTiles = rci->width / rci->tileWidth;
	if( rci->width % rci->tileWidth )
		numTiles++;
	if( rci->height % rci->tileHeight )
		numTiles *= rci->height / rci->tileHeight + 1;
	else
		numTiles *= rci->height / rci->tileHeight;
	
	return numTiles;
}

// this reverses the odd rows in a tile
void RciZigZagTile(RciUns8 *tile, RciUns16 planeDepth, RciUns32 tileWidth, RciUns32 tileHeight)
{
	RciUns8 *cp, ctemp;
	RciUns16 *wp, wtemp;
	RciUns32 row,col;
	RciUns16 planeBytes = planeDepth >> 3;


	row = 1;
	if( planeBytes == 1 )
	{
		cp = tile;
		while( row < tileHeight )
		{
			cp = tile;
			cp += row * tileWidth;

			for(col = 0;col<tileWidth>>1;col++)
			{
				ctemp = cp[col];
				cp[col] = cp[tileWidth-col-1];
				cp[tileWidth-col-1] = ctemp;
			}
			row += 2;
			
		}
	}
	else if( planeBytes == 2 )
	{
		while(row < tileHeight)
		{
			wp = (RciUns16 *)tile;
			wp += row * tileWidth;

			for(col=0;col<tileWidth>>1;col++)
			{
				wtemp = wp[col];
				wp[col] = wp[tileWidth-col-1];
				wp[tileWidth-col-1] = wtemp;
			}
			row += 2;
		}
	}
}
